#include <stdlib.h>
#include <iostream.h>
#include <windows.h>
#include "glut.h"

#include "index.h"
#include "randnum.h"
#include "point.h"
#include "element.h"
#include "grid.h"

// 'grid' has global scope wherever 'grid.h' is included
Grid grid;

//****************************************************************************
// graphics stuff...
//****************************************************************************

/*
 * this does the calculation step
 */
DWORD WINAPI workerProc(void* data)
{
  bool flag = true;

  grid.step();
  flag = grid.nextStep();

  *(bool*)data = flag;

  return 0;
}

/*
 * this drives thru the steps (like the for loop does in the non-graphics version)
 */
static void idleCB()
{
  int waitStatus;
  DWORD workerThread;
  static HANDLE workerHandle = 0;
  static bool keepGoing = true;;

  if (!grid.isDrawing) {
    if (workerHandle == 0) {
      workerHandle = CreateThread(0, 0, workerProc, &keepGoing, 0, &workerThread);
      //cout << "Stepping...\n";
      //cout.flush();
    }
    else {
      waitStatus = WaitForSingleObject(workerHandle, 0);
      
      if (waitStatus == WAIT_OBJECT_0) {
        CloseHandle(workerHandle);
        workerHandle = 0;
        glutPostRedisplay();
      }
      else if (waitStatus == WAIT_ABANDONED) {
        cerr << "mutex problem with worker thread\n";
        exit(-1);
      } 
      else if (waitStatus == WAIT_FAILED) {
        cerr << "something very wrong with the worker thread\n";
        exit(-1);
      }
    }
  }

  if (!keepGoing) {
    glutIdleFunc(NULL);
    cout << "Finished\n";
    cout.flush();
  }
}

/*
 * this function does the drawing
 */
void displayCB()
{
  static bool once = true;

  glClear(GL_COLOR_BUFFER_BIT);

  grid.draw();

  glutSwapBuffers();

  if (once) {
    glutIdleFunc(idleCB);
    once = false;
  }
}

/*
 * press ESC to exit
 */
void keyboardCB(unsigned char key, int, int)
{
  if (key == 27) { // escape key
    exit(0);
  }
}

/*
 * set up the environment for GLUT
 */
void initGraphics(int argc, char* argv[], double xlen, double ylen)
{
  // work out a good size for the window
  double xfrac = 1.0;
  double yfrac = 1.0;

  if (xlen > ylen) {
    yfrac = ylen / xlen;
  }
  else if (xlen < ylen) {
    xfrac = xlen / ylen;
  }

  double maxWindowSide = 750.0; //  pixels

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
  glutInitWindowPosition(50, 50);
  glutInitWindowSize(int(xfrac*maxWindowSide + 0.5), int(yfrac*maxWindowSide + 0.5));
  glutCreateWindow("Finite");
  glutDisplayFunc(displayCB);
  glutKeyboardFunc(keyboardCB);

  glViewport(0, 0, int(xlen), int(ylen));
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0, xlen, 0.0, ylen);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glClearColor(1.0, 1.0, 1.0, 1.0);
}

//****************************************************************************
// end of graphics stuff...
//****************************************************************************

/*
 * recover parameters from command line
 */
bool processCommandLine(int argc, char** argv, int& ncols, int& nrows, int& ntimesteps)
{
  if (argc < 4) {
    return false;
  }

  ncols = atoi(argv[1]);
  nrows = atoi(argv[2]);
  ntimesteps = atoi(argv[3]);

  return true;
}

/*
 * entry point
 */
int main(int argc, char* argv[])
{
  int ncols, nrows, ntimeSteps;

  ncols = nrows = -1;

  // find size of grid
  if (!processCommandLine(argc, argv, ncols, nrows, ntimeSteps)) {
    cerr << "usage: " << argv[0] << " <number columns> <number rows> <number timesteps>\n";
    exit(-1);
  }

  // check the input
  if (ncols < 1 || nrows < 1) {
    cerr << argv[0] << ": numbers of columns and rows must be positive\n";
    exit(-1);
  }

  // create the grid object (and generate the 2D array of elements within it)
  grid = Grid(ncols, nrows, ntimeSteps);

  // initialise elements with random (u,v) field
  grid.init();

  // set up the graphics
  initGraphics(argc, argv, double(ncols), double(nrows));

  // hand over to GLUT library
  glutMainLoop();

  return 0;
}

/*
 * E N D  O F  F I L E
 */
